/**
 * Copyright (C) 1998 Stewart Allen (stewart@neuron.com)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

import java.io.*;
import java.net.*;
import java.util.*;

public class RPC2_Packet
{
	RPC2_BodyStruct bodyStruct;
	ByteArrayOutputStream dataOut;
	byte data[];
	int len;
	int pos;

	int ProtoVersion;	// always 8 or something is broken
	int RemoteHandle;
	int LocalHandle;
	int Flags;			// b[0]=se, b[1]=color, b[2,3]=rpc
	int BodyLength;
	int SeqNumber;		// odd=client, even=server
	int Opcode;			// >0=sys-specific, <0=reserved
	int SEFlags;
	int SEDataOffset;
	int SubsysId;
	int ReturnCode;		// meaningless on request packets
	int TimeStamp;		// for rpc timing
	int BindTime;		// ???
	int Lamport;		// ??? - global unique # for this field
	int Uniquefier;		// for Init1 packets - must match NewConnection

	static int nextLamport = 1;
	static byte[] zeros = { '\0', '\0', '\0', '\0' };

	final static int bodyOffset = 15 * 4;	// from 15 "int" fields

	// for outbound packets - gets settings from Connection
	public RPC2_Packet(Connection conn)
	{
		this();
		this.RemoteHandle = conn.RemoteHandle;
		this.LocalHandle = conn.LocalHandle;
		this.Uniquefier = conn.hashCode();
		this.SeqNumber = conn.NextSeqNumber;
		this.SubsysId = conn.SubsysId;
		this.TimeStamp = conn.getTimeStamp();
	}

	// for outbound packets - reverses inbound handles
	// this is for "response" packets inside "Servers"
	public RPC2_Packet(RPC2_Packet in, Connection conn)
	{
		this();
		this.RemoteHandle = in.LocalHandle;
		this.LocalHandle = in.RemoteHandle;
		this.Uniquefier = in.Uniquefier;
		this.SubsysId = in.SubsysId;
		this.SeqNumber = conn.NextSeqNumber;
		this.TimeStamp = conn.getTimeStamp();
	}

	// for outbound packets
	public RPC2_Packet()
	{
		ProtoVersion = RPC2.RPC2_PROTOVERSION;
		Lamport = nextLamport();
	}

	synchronized int nextLamport()
	{
		return nextLamport++;
	}

	// for inbound packets
	public RPC2_Packet(byte data[])
	{
		setData(data);

		ProtoVersion = getInt();	// 0
		RemoteHandle = getInt();	// 1
		LocalHandle = getInt();		// 2
		Flags = getInt();			// 3
		BodyLength = getInt();		// 4
		SeqNumber = getInt();		// 5
		Opcode = getInt();			// 6
		SEFlags = getInt();			// 7
		SEDataOffset = getInt();	// 8
		SubsysId = getInt();		// 9
		ReturnCode = getInt();		// 10
		Lamport = getInt();			// 11
		Uniquefier = getInt();		// 12
		TimeStamp = getInt();		// 13
		BindTime = getInt();		// 14

	}

	// -- Byte buffer get/set position --
	
	public void setPos(int pos)
	{
		this.pos = pos;
	}

	public int getPos()
	{
		return pos;
	}

	public void setData(byte data[])
	{
		this.len = data.length;
		this.data = data;
		this.pos = 0;
	}

	// -- Put values into a byte array --

	public void put(RPC2_BodyStruct struct)
	{
		struct.encode(this);
	}

	public void put(long val)
	{
		put((int)val);
	}

	public void put(int val)
	{
		for (int i=0; i<4; i++)
			dataOut.write( (val >> ((3-i)*8)) & 0xff );
	}

	public void put(String s, boolean padded)
	{
		put(s.getBytes(), padded);
	}

	// puts a byte array
	// optionally padded to 4 byte boundary
	public void put(byte in[], boolean padded)
	{
		if (in == null || in.length == 0)
			return;
		try {
			dataOut.write(in,0,in.length);
			if (padded)
				dataOut.write(zeros, 0, pos % 4);
		} catch (Exception e) { e.printStackTrace(); }
	}

	public void setBody(RPC2_BodyStruct body)
	{
		// store local pointer
		this.bodyStruct = body;
	}

	public void markSEDataOffset()
	{
		SEDataOffset = dataOut.size();
	}

	public byte[] getPacketData()
	{
		if (data != null)
			return data;

		// create body byte array
		dataOut = new ByteArrayOutputStream(300);

		// add body struct if not null
		if (bodyStruct != null)
			bodyStruct.encode(this);

		// set Body length from acculumated data
		BodyLength = dataOut.size();

		// store current buffer
		byte bdata[] = dataOut.toByteArray();

		// create header byte array
		dataOut = new ByteArrayOutputStream(120);
		put(ProtoVersion);
		put(RemoteHandle);
		put(LocalHandle);
		put(Flags);
		put(BodyLength);
		put(SeqNumber);
		put(Opcode);
		put(SEFlags);
		put(SEDataOffset);
		put(SubsysId);
		put(ReturnCode);
		put(Lamport);
		put(Uniquefier);
		put(TimeStamp);
		put(BindTime);

		// append body and sedata
		try { dataOut.write(bdata); } catch (Exception e) { e.printStackTrace(); }

		// set this.data to byte array
		setData(dataOut.toByteArray());

		return data;
	}

	// -- Get values from byte buffer --

	public byte getOneByte()
	{
		return data[pos++];
	}

	// RPC2_Byte
	public byte getByte()
	{
		byte b = data[pos++];
		// skip three padding bytes
		pos += 3;
		return b;
	}

	// RPC2_Integer
	public int getInt()
	{
		int val = 0;
		for (int i=0; i<4; i++)
			val |= ( (data[pos+i] & 0xff) << ((3-i)*8) );
		pos += 4;
		return val;
	}

	public long getLong()
	{
		long val = 0;
		for (int i=0; i<8; i++)
			val |= ( (data[pos+i] & 0xff) << ((7-i)*8) );
		pos += 8;
		return val;
	}

	public byte[] getBytes(int getLen)
	{
		if (pos > len || pos+getLen > len)
			return new byte[0];

		// add padding out to nearest 4 byte boundary if necessary
		int add = getLen % 4;
		if (add != 0)
			add = 4 - add;
		//System.err.println("pos = "+pos+" getLen = "+getLen+" add = "+add);

		byte d[] = new byte[getLen];
		System.arraycopy(data,pos,d,0,getLen);
		pos += (getLen + add);
		return d;
	}

	// RPC2_CountedBS as java.lang.String
	public String getCountedBString()
	{
		byte b[] = getCountedBS();
		if (b == null)
			return null;
		return new String(b);
	}

	// RPC2_CountedBS
	public byte[] getCountedBS()
	{
		int len = getInt();
		return getBytes(len);
	}

	// RPC2_String
	public String getString()
	{
		int len = getInt();
		byte b[] = getBytes(len+1);
		return new String(b, 0, len);
	}

	public String getNullTermString()
	{
		int start = pos;
		while (data[pos++] != 0)
			;

		int len = pos - start - 1;
		byte ndata[] = new byte[len];

		System.arraycopy(data, start, ndata, 0, len);

		return new String(ndata);
	}

	// -- DEBUGGING utilities ----------

	public void print()
	{
		getPacketData();

		byte b[] = data;
		char ch[] = new char[54];
		int is[] = new int[3];

		for (int i=0; i<54; i++)
			ch[i] = ' ';

		Util.log("-- PACKET ("+len+") --");
		for (int i=0; i<len; i+= 12)
		{
			for (int j=0; j<3; j++)
			{
				is[j] = 0;
				for (int k=0; k<4; k++)
				{
					int inc = j*4+k;
					int pos = i+inc;

//					if (pos >= len)
//						break;

					int p2 = inc*3;
					p2 += p2/12;
					if (pos >= len) {
						ch[p2] = '.';
						ch[p2+1] = '.';
						ch[40+inc] = ' ';
						is[j] = -1;
					} else {
						ch[p2] = Character.forDigit((b[pos] >> 4) & 0xF, 16);
						ch[p2+1] = Character.forDigit(b[pos] & 0xF, 16);

						char cch = (char)b[pos];
						if (cch >= 32 && cch <= 126)
							ch[40+inc] = cch;
						else
							ch[40+inc] = '_';

						is[j] |= ( (cch & 0xff) << ((3-k)*8) );
					}
				}
			}
			Util.log(
				new String(ch) + " " + intPad(i,3) + "  " +
				intPad(is[0]) + " " + intPad(is[1]) + " " + intPad(is[2])
			);
		}
	}

	public String intPad(int x)
	{
		return intPad(x, 10);
	}

	public String intPad(int x, int w)
	{
		StringBuffer sb = new StringBuffer(Integer.toString(x));
		while (sb.length() < w)
			sb.append(' ');
		return sb.toString();
	}

	public String toString()
	{
		getPacketData();
		String opName = RPC2.getOpcodeName(Opcode);
		return	"<RPC2 Packet> " + (opName != null ? opName : "unknown") +
				"\n   proto   =   " + intPad(ProtoVersion) +
				"     remote  =   " + intPad(RemoteHandle) +
				"     local   =   " + intPad(LocalHandle) +
				"\n   flags   =   " + RPC2_Util.getHex(Flags) + " " +
				"     body    =   " + intPad(BodyLength) +
				"     seqNum  =   " + intPad(SeqNumber) +
				"\n   opCode  =   " + intPad(Opcode) +
				"     SEFlags =   " + intPad(SEFlags) +
				"     SEOffst =   " + intPad(SEDataOffset) +
				"\n   subsys  =   " + intPad(SubsysId) +
				"     return  =   " + intPad(ReturnCode) +
				"     lamport =   " + intPad(Lamport) +
				"\n   unique  =   " + intPad(Uniquefier) +
				"     time    =   " + intPad(TimeStamp) +
				"     bind    =   " + intPad(BindTime);
	}
}


